!***********************************************************************
!
!    Copyright (c) 2013, Lawrence Livermore National Security, LLC.
!                        Produced at the Lawrence Livermore National
!                        Laboratory.
!                        Written by Nicolas Schunck, schunck1@llnl.gov
!
!    LLNL-CODE-470611 All rights reserved.
!
!    Copyright 2011, N. Schunck, J. Dobaczewski, J. McDonnell,
!                    W. Satula, J.A. Sheikh, A. Staszczak,
!                    M. Stoitsov, P. Toivanen
!    Copyright 2009, J. Dobaczewski, W. Satula, B.G. Carlsson, J. Engel,
!                    P. Olbratowski, P. Powalowski, M. Sadziak,
!                    J. Sarich, N. Schunck, A. Staszczak, M. Stoitsov,
!                    M. Zalewski, H. Zdunczuk
!    Copyright 2004, 2005, J. Dobaczewski, P. Olbratowski
!    Copyright 1997, 2000, J. Dobaczewski, J. Dudek
!
!    This file is part of HFODD.
!
!    HFODD is free software: you can redistribute it and/or modify it
!    under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    HFODD is distributed in the hope that it will be useful, but
!    WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with HFODD. If not, see <http://www.gnu.org/licenses/>.
!
!    OUR NOTICE AND TERMS AND CONDITIONS OF THE GNU GENERAL PUBLIC
!    LICENSE
!
!    Our Preamble Notice
!
!      A. This notice is required to be provided under our contract
!         with the U.S. Department of Energy (DOE). This work was
!         produced at the Lawrence Livermore National Laboratory under
!         Contract No. DE-AC52-07NA27344 with the DOE.
!      B. Neither the United States Government nor Lawrence Livermore
!         National Security, LLC nor any of their employees, makes any
!         warranty, express or implied, or assumes any liability or
!         responsibility for the accuracy, completeness, or usefulness
!         of any information, apparatus, product, or process disclosed,
!         or represents that its use would not infringe privately-owned
!         rights.
!      C. Also, reference herein to any specific commercial products,
!         process, or services by trade name, trademark, manufacturer
!         or otherwise does not necessarily constitute or imply its
!         endorsement, recommendation, or favoring by the United States
!         Government or Lawrence Livermore National Security, LLC. The
!         views and opinions of authors expressed herein do not
!         necessarily state or reflect those of the United States
!         Government or Lawrence Livermore National Security, LLC, and
!         shall not be used for advertising or product endorsement
!         purposes.
!
!    The precise terms and conditions for copying, distribution and
!    modification are contained in the file COPYING.
!
!-------------------------------------------------------------------!
!                                                                   !
!         MASS PARAMETERS OF 1/2(Q_LM + Q*_LM) OPERATORS            !
!                                                                   !
! THIS MODULE CONSISTS OF TWO MAIN ROUTINES:                        !
!     1. MASPAR() - A) MASS PARAMETERS AT A GIVEN CONSTRAINTS       !
!                   B) ZERO POINT ENERGIES E0 (GCM and CRANKING)    !
!     2. ROTCOR() - A) MOMENTS OF INERTIA Jx, Jy, Jz,               !
!                   B) ROTATIONAL ENERGY CORRECTIONS Ex, Ey, Ez     !
!                                                                   !
! CALLING CONVENTION:                                               !
!                                                                   !
!      CALL MASPAR(TERMNT,ICHARG,MREVER)                            !
!                                                                   !
!      CALL ROTCOR(TERMNT,ICHARG,MREVER,XMOMFC,YMOMFC,ZMOMFC,       !
!     *                                 NXMAXX,NYMAXX,NZMAXX)       !
!                                                                   !
!-------------------------------------------------------------------!

 MODULE hfodd_inertia

     USE hfodd_sizes

     USE MAT_PP
     USE PAIDEL
     USE ALLWAV
     USE ALLQUA

     INTEGER, PARAMETER :: NDEFOR=NDMULT, NDCOMP=(NDEFOR*(NDEFOR+1))/2, NDEMOM=3

     INTEGER, PUBLIC, SAVE :: NOCOMP,NOMULT,IIPAIR
     INTEGER, PUBLIC, SAVE :: INDMAS(NDCOMP,4)

     DOUBLE PRECISION, PUBLIC, SAVE :: ATDMAST(NDCOMP),ATDMAS(NDCOMP,0:2)
     DOUBLE PRECISION, PUBLIC, SAVE :: CRAMAS(NDCOMP,0:2),GCMMAS(NDCOMP,0:2)
     DOUBLE PRECISION, PUBLIC, SAVE :: E0_GCM(0:2), E0_ATD(0:2)
     DOUBLE PRECISION, PUBLIC, SAVE :: COVMET(NDEFOR,NDEFOR,0:2)
     DOUBLE PRECISION, PUBLIC, SAVE :: EMOMET(NDEFOR,NDEFOR,NDEMOM)
     DOUBLE PRECISION, PUBLIC, SAVE :: P1(NDEFOR,NDEFOR,0:1),P2(NDEFOR,NDEFOR,0:1)
     DOUBLE PRECISION, PUBLIC, SAVE :: SK1(NDCOMP,0:2),SK3(NDCOMP,0:2)

 CONTAINS

     !-------------------------------------------------------------------!
     ! Initialization of all public variables                            !
     !-------------------------------------------------------------------!
     SUBROUTINE INIMAS()

       EMOMET=0.0D0
       COVMET=0.0D0
       ATDMAS=0.0D0
       CRAMAS=0.0D0
       GCMMAS=0.0D0
       E0_GCM=0.0D0
       E0_ATD=0.0D0

       INDMAS=0

       NOCOMP=0
       NOMULT=0

     END SUBROUTINE INIMAS

     !-------------------------------------------------------------------!
     !                                                                   !
     !  This subroutine calculates the inertia tensors on the basis of   !
     !     1. Adiabatic Time Dependent Hartree-Fock formula .... ATDHF   !
     !     2. Generator Coordinate Method formula (GOA) .......... GCM   !
     !     3. Cranking Model Formula ........................ CRANKING   !
     !                                                                   !
     !  Remarks:                                                         !
     !     - the components of mass with lambda = 0 and 1 are not        !
     !       calculated. The minimal multipolarity is lambda = 2.        !
     !     - the routine MAS_PR should be called in the "neutron" sector !
     !       (ICHARG=0) and in the "proton" sector (ICHARG=1) of HFODD   !
     !       after solving the HFB equations for a given type of         !
     !       particles.                                                  !
     !                                                                   !
     !  References:                                                      !
     !     1. Pomorski, et al., Nucl. Phys. A ...                        !
     !     2. Gozdz, et al., Nucl. Phys. A 442, (1985)                   !
     !     3. Libert, et al., Phys. Rev. C 60, 054301 (1999)             !
     !     4. Goutte, et al., Phys. Rev. C 71, 024316 (2005)             !
     !                                                                   !
     !  Inputs                                                           !
     !                                                                   !
     !
     !
     !  Outputs                                                          !
     !                                                                   !
     !     The values of the mass components are stored in the matrix    !
     !     ATDMAS(NCOUNT,ISOSPI). The first index NCOUNT refers to the   !
     !     pair (i,j) of the deformations, i=(l_i, m_i), j=(l_j, m_j),   !
     !     the second index to the contribution from protons, neutrons   !
     !     or both. The correspondence between pair index and actual     !
     !     multipolarities is obtained by the vector INDMAS(:,:) in the  !
     !     following way:                                                !
     !                                                                   !
     !       INDMAS(NCOUNT,1) <-- LAMBDA,  INDMAS(NCOUNT,2) <-- MIUA     !
     !       INDMAS(NCOUNT,3) <-- LAMBDB,  INDMAS(NCOUNT,4) <-- MIUB     !
     !                                                                   !
     !         ATDHF  | cranking |  GCM   (indices)      meaning         !
     !         --------------------------------------------------        !
     !         ATDMAS |  CRAMAS  | GCMMAS (ncount,0) <--  protons        !
     !         ATDMAS |  CRAMAS  | GCMMAS (ncount,1) <-- neutrons        !
     !         ATDMAS |  CRAMAS  | GCMMAS (ncount,2) <-- total-pn        !
     !                                                                   !
     !     The number of actual independent components of the mass       !
     !     tensor is returned in NOCOMP.                                 !
     !                                                                   !
     !  Meaning of some parameters                                       !
     !                                                                   !
     !     NDEFOR      -  Maximal number of deformations defining the    !
     !                    collective space for this module. Can not be   !
     !                    greater than NDMULT                            !
     !     NDMULT      -  Maximal number of constraints in HFODD         !
     !     NOMULT      -  Actual number of constraints in HFODD          !
     !                                                                   !
     !     NDCOMP      -  Maximum number of independent components of    !
     !                    the inertia tensor, NDEFOR*(NDEFOR+1))/2       !
     !     NOCOMP      -  Actual number of independent components        !
     !                                                                   !
     !     NDEMOM      -  Number of calculated energy moments. It is     !
     !                    fixed to 3 by default                          !
     !                                                                   !
     !     ATDMAS(:,:) -  Independent components of the mass parameter   !
     !     CRAMAS(:,:)    in the case of ATDHF (ATD), cranking (CRA) and !
     !     GCMMAS(:,:)    GCM (GCM) case.                                !
     !                                                                   !
     !     INDMAS(:,:) -  Multi-index mapping pairs of indexes in the    !
     !                    mass tensor with actual multipolarities        !
     !                                                                   !
     !     EMOMET      -  Energy moments (internal) M_{-k}^{ij}          !
     !                    dimension(NDEFOR,NDEFOR,NDEMOM)                !
     !                                                                   !
     !-------------------------------------------------------------------!
     SUBROUTINE MAS_PR(TERMNT,ICHARG,MREVER,IF2FIN,IF2FIP,L_COLL,M_COLL,N_COLL,E0CORR)

       LOGICAL :: TERMNT,CHEKIT

       REAL, DIMENSION (NDEFOR,NDEFOR) :: AUX1,AUX2,AUX3

       INTEGER, DIMENSION(:), pointer :: L_COLL, M_COLL

       ! Inputs
       COMMON &
              /CFIPRI/ NFIPRI
       COMMON                                           &
              /QCNSTR/ STIFFQ(0:NDMULT,-NDMULT:NDMULT), &
                       QASKED(0:NDMULT,-NDMULT:NDMULT), &
                       IFLAGQ(0:NDMULT,-NDMULT:NDMULT)
       COMMON &
              /CAOCNP/ V2OCNP(1:2*NDBASE,0:NDISOS)

       ! Internal
       DATA &
            NIMULT /-1/, CHEKIT /.FALSE./

       !-------------------------------------------------------------------!
       !  Handles return cases                                             !
       !-------------------------------------------------------------------!

       IF (.NOT.TERMNT) RETURN

       ! Return if no constraints are set, or no collective space has been
       ! preset by the user
       IF (NIMULT.EQ.0.OR.N_COLL.EQ.0) RETURN

       CALL CPUTIM('MAS_PR',1)

       !-------------------------------------------------------------------!
       !  Define the collective space as the number of active constraints  !
       !-------------------------------------------------------------------!

       IF (CHEKIT) THEN

           NOCOMP=0

           ! Number of LAMBDA>1 constraints
           NIMULT=0

           ! Number of different components with LAMBDA>1
           NOMULT=0
           MINLAM=2

           DO LAMBDA=MINLAM,NDMULT
              DO MIUA=-LAMBDA,LAMBDA

                 IF(IFLAGQ(LAMBDA,MIUA).EQ.1) THEN
                    NIMULT=NIMULT+1

                    DO LAMBDB=MINLAM,LAMBDA
                       DO MIUB=-LAMBDB,MIUA

                          IF(IFLAGQ(LAMBDB,MIUB).EQ.1) THEN

                             NOCOMP=NOCOMP+1
                             INDMAS(NOCOMP,1)=LAMBDA
                             INDMAS(NOCOMP,2)=MIUA
                             INDMAS(NOCOMP,3)=LAMBDB
                             INDMAS(NOCOMP,4)=MIUB

                          END IF

                       END DO
                    END DO

                 END IF

              END DO
           END DO

           ! Number of multipoles
           NOMULT=NIMULT

           ! Check already done
           CHEKIT=.FALSE.

       !-------------------------------------------------------------------!
       !  The collective space has been preset in the input file           !
       !-------------------------------------------------------------------!

       ELSE

           NOCOMP=0

           ! Number of LAMBDA>1 collective variables
           NOMULT=N_COLL

           ! Number of different components with LAMBDA>1
           DO ICOLLA=1,N_COLL
              DO ICOLLB=ICOLLA,N_COLL

                 LAMBDA=L_COLL(ICOLLA)
                 MIUA  =M_COLL(ICOLLA)

                 LAMBDB=L_COLL(ICOLLB)
                 MIUB  =M_COLL(ICOLLB)

                 NOCOMP=NOCOMP+1

                 INDMAS(NOCOMP,1)=LAMBDA
                 INDMAS(NOCOMP,2)=MIUA
                 INDMAS(NOCOMP,3)=LAMBDB
                 INDMAS(NOCOMP,4)=MIUB

              END DO
           END DO

       END IF

       ! Jhilam: Collective variables are lambda2 parameters of Lipkin-Nogami
       IF(IIPAIR.EQ.1)THEN
          DO I=1,N_COLL
             J=(I*(I+1))/2
             NOCOMP=NOCOMP+1
             INDMAS(NOCOMP,1)=INDMAS(J,3)
             INDMAS(NOCOMP,2)=INDMAS(J,4)
          ENDDO
          NOCOMP=NOCOMP+1
          NIMULT=NIMULT+1
          NOMULT=NIMULT
       END IF

       ! Check that dimensions are OK
       IF (NOMULT.GT.NDEFOR) STOP 'MASPAR: INCREASE NDEFOR PARAMETER!'

       !-------------------------------------------------------------------!
       !         ENERGY MOMENTS N=1..NDEMOM FOR ALL MASS COMPONENTS
       !-------------------------------------------------------------------!

       ! Calculate energy moments (EMOMET)
       CALL INVEMO(MREVER,ICHARG)

       !-------------------------------------------------------------------!
       !
       !  ATTENTION!!!
       !
       !     Factor (1/2) in the definitions of sigma^(k) comes from the
       !     difference in the way we calculate moments m^k  where summs
       !     run over the whole set of indices mu,nu and in sigma (where
       !     the sums run over the positive indices only).
       !
       !     This assumes the following time-reversal property of the
       !     collective variable operators q: |<0|q|-m-n>|=|<0|q|mn>|
       !
       !-------------------------------------------------------------------!

       P1(:,:,ICHARG)=EMOMET(:,:,1)
       P2(:,:,ICHARG)=EMOMET(:,:,2)
       ! Mass parameters B_{ij} given by Eq.(23), Libert et al. (1999)
       ! M-1**-1 * M-3 * M-1**-1 --> AUX1
       AUX1=EMOMET(:,:,1)
       CALL S2VCPY(AUX1,NOMULT,NDEFOR,SK1(1,ICHARG),NOCOMP,NDCOMP)
       AUX2=EMOMET(:,:,3)
       CALL S2VCPY(AUX2,NOMULT,NDEFOR,SK3(1,ICHARG),NOCOMP,NDCOMP)
       !
       IF(ICHARG.EQ.1) THEN
         ! SIG2 =  M1^(-1) M2 M1^(-1) --> AUX2
         AUX2=P1(:,:,0)+P1(:,:,1)
         AUX1=P2(:,:,0)+P2(:,:,1)
         CALL INVMUL(AUX2,AUX1,AUX3,NOMULT,NDEFOR)
         ! Save metric (SIG2)
         COVMET(:,:,2)=AUX2
         ! E0_GCM = 1/4 Tr( Sig2^(-1) Sig1 ) (SEE: Pomorski NP, 1989) (16)
         ! Get metric (SIG2) and invert it
         AUX2=COVMET(:,:,2)
         CALL INVERT(AUX2,AUX3,NOMULT,NDEFOR)
          ! Make S1=1/4 M1^-1
         AUX1=P1(:,:,0)+P1(:,:,1)
         CALL INVERT(AUX1,AUX3,NOMULT,NDEFOR)
         ! Make SIG2^-1*SIG1
         AUX3=MATMUL(AUX2,AUX1)
         ! Get trace of 1/4*(SIG2^-1 SIG1)
         E0_GCM(2)=0.25*TRACEM(AUX3,NOMULT,NDEFOR)
         AUX1=0.0D0
         AUX2=0.0D0
         SK1(:,2)=SK1(:,0)+SK1(:,1)
         SK3(:,2)=SK3(:,0)+SK3(:,1)
         CALL V2SCPY(SK1(1,2),NOCOMP,NDCOMP,AUX1,NOMULT,NDEFOR)
         CALL V2SCPY(SK3(1,2),NOCOMP,NDCOMP,AUX3,NOMULT,NDEFOR)
         CALL INVMUL(AUX1,AUX3,AUX2,NOMULT,NDEFOR)
         CALL S2VCPY(AUX1,NOMULT,NDEFOR,SK3(1,2),NOCOMP,NDCOMP)
         ATDMAS=SK3
       END IF

       E0CORR=E0_GCM(2)

       ATDMAST(:)=0.0D0
       IF(ICHARG.EQ.1)THEN
          DO J=1,NOCOMP
             ATDMAST(J)=ATDMAS(J,2)
          END DO
          N_COMP=NDCOMP
       ENDIF

       !-------------------------------------------------------------------!
       !     PRINTING MASS TENSOR
       !-------------------------------------------------------------------!

       IF (ICHARG.EQ.1) THEN
           CALL PRIMAS(E0CORR)
       END IF

       CALL CPUTIM('MAS_PR',0)

       RETURN
     END SUBROUTINE MAS_PR

     !--------------------------------------------------------------------!
     !  THis routine adds n and p and/or writes mass tensor components    !
     !                                                                    !
     !  Meaning of some parameters                                        !
     !    ICHARG - Standard 0, 1 = N, P; WRITE                            !
     !                         2 = add, form total components and write   !
     !    NOMULT - Number of constraints (LAMBDA>=2)                      !
     !    NOCOMP - Total number of different components of mass tensors   !
     !                                                                    !
     !--------------------------------------------------------------------!
     SUBROUTINE PRIMAS(E0CORR)

       COMMON /CFIPRI/ NFIPRI

       CHARACTER(LEN=8) NAMEPN
       CHARACTER(LEN=9) MEVBCH
       CHARACTER(LEN=2) DIVBY2,EMPTCH,OHAFCH

       !  Particle type
       NAMEPN='   TOTAL'

       !  Mass parameter units
       OHAFCH='/2'
       EMPTCH='  '
       MEVBCH='1/MeV b**'

       WRITE(NFIPRI,'(79(''*''),/,''*'',77X,''*'')')
       WRITE(NFIPRI,'(''*'',29X,''COLLECTIVE INERTIA'',30X,''*'')')
       WRITE(NFIPRI,'(''*'',77X,''*'')')
       WRITE(NFIPRI,'(''*  MASS TENSOR B_{ij}/HbarC**2 WITH i=(L,M) AND j=(LP,MP)'',8X,A8,5X,''*'')') NAMEPN
       WRITE(NFIPRI,'(''*'',77X,''*'')')
       WRITE(NFIPRI,'(''*    L   M    LP  MP       ATDHF       UNITS      '',28X,''*'')')

       DO ICOMPS=1,NOCOMP

          IUNITS=INDMAS(ICOMPS,1)+INDMAS(ICOMPS,3)
          DIVBY2=OHAFCH

          IF (MOD(IUNITS,2).EQ.0) THEN
              DIVBY2=EMPTCH
              IUNITS=IUNITS/2
          ELSE
              DIVBY2=OHAFCH
          END IF

          WRITE(NFIPRI,'(''*'',1X,2I4,2X,2I4,1E14.4,2X,A9,I1,A2,2X,''*'')') &
                INDMAS(ICOMPS,1),INDMAS(ICOMPS,2), &
                INDMAS(ICOMPS,3),INDMAS(ICOMPS,4), &
                ATDMAST(ICOMPS),MEVBCH,IUNITS,DIVBY2

       END DO

       ! Zero point energy
       WRITE(NFIPRI,'(''*'',77X,''*'')')
       WRITE(NFIPRI,'(''*  ZERO-POINT ENERGY CORRECTION - ATDHF:'',F9.4,'' MeV'',25X,''*'')') E0CORR
       WRITE(NFIPRI,'(''*'',77X,''*'',/,79(''*''),/)')

       RETURN
     END SUBROUTINE PRIMAS

     !-------------------------------------------------------------------!
     !                                                                   !
     !  This subroutine calculates the inertia tensors on the basis of   !
     !     1. Adiabatic Time Dependent Hartree-Fock formula .... ATDHF   !
     !     2. Generator Coordinate Method formula (GOA) .......... GCM   !
     !     3. Cranking Model Formula ........................ CRANKING   !
     !                                                                   !
     !  Remarks:                                                         !
     !     - the components of mass with lambda = 0 and 1 are not        !
     !       calculated. The minimal multipolarity is lambda = 2.        !
     !     - the routine MASPAR should be called in the "neutron" sector !
     !       (ICHARG=0) and in the "proton" sector (ICHARG=1) of HFODD   !
     !       after solving the HFB equations for a given type of         !
     !       particles.                                                  !
     !                                                                   !
     !  References:                                                      !
     !     1. Pomorski, et al., Nucl. Phys. A ...                        !
     !     2. Gozdz, et al., Nucl. Phys. A 442, (1985)                   !
     !     3. Libert, et al., Phys. Rev. C 60, 054301 (1999)             !
     !     4. Goutte, et al., Phys. Rev. C 71, 024316 (2005)             !
     !                                                                   !
     !  Inputs                                                           !
     !                                                                   !
     !
     !
     !  Outputs                                                          !
     !                                                                   !
     !     The values of the mass components are stored in the matrix    !
     !     ATDMAS(NCOUNT,ISOSPI). The first index NCOUNT refers to the   !
     !     pair (i,j) of the deformations, i=(l_i, m_i), j=(l_j, m_j),   !
     !     the second index to the contribution from protons, neutrons   !
     !     or total. The correspondence between pair index and actual    !
     !     multipolarities is obtained by the vector INDMAS(:,:) in the  !
     !     following way:                                                !
     !                                                                   !
     !       INDMAS(NCOUNT,1) <-- LAMBDA,  INDMAS(NCOUNT,2) <-- MIUA     !
     !       INDMAS(NCOUNT,3) <-- LAMBDB,  INDMAS(NCOUNT,4) <-- MIUB     !
     !                                                                   !
     !         ATDHF  | cranking |  GCM   (indices)      meaning         !
     !         --------------------------------------------------        !
     !         ATDMAS |  CRAMAS  | GCMMAS (ncount,0) <--  protons        !
     !         ATDMAS |  CRAMAS  | GCMMAS (ncount,1) <-- neutrons        !
     !         ATDMAS |  CRAMAS  | GCMMAS (ncount,2) <-- total-pn        !
     !                                                                   !
     !     The number of actual independent components of the mass       !
     !     tensor is returned in NOCOMP.                                 !
     !                                                                   !
     !  Meaning of some parameters:                                      !
     !                                                                   !
     !     NDEFOR      -  Maximal number of deformations defining the    !
     !                    collective space for this module. Can not be   !
     !                    greater than NDMULT                            !
     !     NDMULT      -  Maximal number of constraints in HFODD         !
     !     NOMULT      -  Actual number of constraints in HFODD          !
     !                                                                   !
     !     N_COLL      -  Number of collective variables: can be either  !
     !                    set in HFODD input file, or generated "on the  !
     !                    fly                                          	 !
     !     NDCOMP      -  Maximum number of independent components of    !
     !                    the inertia tensor, NDEFOR*(NDEFOR+1))/2       !
     !     NOCOMP      -  Actual number of independent components        !
     !                                                                   !
     !     NDEMOM      -  Number of calculated energy moments. It is     !
     !                    fixed to 3 by default                          !
     !                                                                   !
     !     ATDMAS(:,:) -  Independent components of the mass parameter   !
     !     CRAMAS(:,:)    in the case of ATDHF (ATD), cranking (CRA) and !
     !     GCMMAS(:,:)    GCM (GCM) case.                                !
     !                                                                   !
     !     INDMAS(:,:) -  Multi-index mapping pairs of indexes in the    !
     !                    mass tensor with actual multipolarities        !
     !                                                                   !
     !     EMOMET      -  Energy moments (internal) M_{-k}^{ij}          !
     !                    dimension(NDEFOR,NDEFOR,NDEMOM)                !
     !                                                                   !
     !-------------------------------------------------------------------!
     SUBROUTINE MASPAR(TERMNT,ICHARG,MREVER,L_COLL,M_COLL,N_COLL)

       LOGICAL TERMNT,CHEKIT

       REAL, DIMENSION (NDEFOR,NDEFOR) :: AUX1,AUX2,AUX3

       INTEGER, DIMENSION(:), pointer :: L_COLL, M_COLL

       COMMON                                           &
              /QCNSTR/ STIFFQ(0:NDMULT,-NDMULT:NDMULT), &
                       QASKED(0:NDMULT,-NDMULT:NDMULT), &
                       IFLAGQ(0:NDMULT,-NDMULT:NDMULT)
       COMMON /CFIPRI/ NFIPRI

       ! Internal
       DATA &
            NIMULT /-1/, CHEKIT /.FALSE./

       !-------------------------------------------------------------------!
       !  Handles return cases                                             !
       !-------------------------------------------------------------------!

       IF (.NOT.TERMNT) RETURN

       ! Return if no constraints are set, or no collective space has been
       ! preset by the user
       IF (NIMULT.EQ.0.OR.N_COLL.EQ.0) RETURN

       !-------------------------------------------------------------------!
       !  Define the collective space as the number of active constraints  !
       !-------------------------------------------------------------------!

       IF (CHEKIT) THEN

           NOCOMP=0

           ! Number of LAMBDA>1 constraints
           NIMULT=0

           ! Number of different components with LAMBDA>1
           NOMULT=0
           MINLAM=2
           N_COLL=0

           DO LAMBDA=MINLAM,NDMULT
              DO MIUA=-LAMBDA,LAMBDA
                 N_COLL=N_COLL+1

                 IF(IFLAGQ(LAMBDA,MIUA).EQ.1) THEN
                    NIMULT=NIMULT+1

                    DO LAMBDB=MINLAM,LAMBDA
                       IF(LAMBDA.EQ.2)THEN
                          MIUX=MIUA
                       ELSE
                          MIUX=LAMBDB
                       ENDIF
                       DO MIUB=-LAMBDB,MIUX

                          IF(IFLAGQ(LAMBDB,MIUB).EQ.1) THEN

                             NOCOMP=NOCOMP+1
                             INDMAS(NOCOMP,1)=LAMBDA
                             INDMAS(NOCOMP,2)=MIUA
                             INDMAS(NOCOMP,3)=LAMBDB
                             INDMAS(NOCOMP,4)=MIUB

                          END IF

                       END DO
                    END DO

                 END IF

              END DO
           END DO

           ! Number of multipoles
           NOMULT=NIMULT

           ! Dimensions
           IF (NOMULT.GT.NDEFOR) STOP 'MASPAR: INCREASE NDEFOR PARAMETER!'

           ! Check already done
           CHEKIT=.FALSE.

       !-------------------------------------------------------------------!
       !  The collective space has been preset in the input file           !
       !-------------------------------------------------------------------!

       ELSE

           NOCOMP=0

           ! Number of LAMBDA>1 collective variables
           NOMULT=N_COLL

           ! Number of different components with LAMBDA>1
           DO ICOLLA=1,N_COLL
              DO ICOLLB=ICOLLA,N_COLL

                 LAMBDA=L_COLL(ICOLLA)
                 MIUA  =M_COLL(ICOLLA)

                 LAMBDB=L_COLL(ICOLLB)
                 MIUB  =M_COLL(ICOLLB)

                 NOCOMP=NOCOMP+1

                 INDMAS(NOCOMP,1)=LAMBDA
                 INDMAS(NOCOMP,2)=MIUA
                 INDMAS(NOCOMP,3)=LAMBDB
                 INDMAS(NOCOMP,4)=MIUB

              END DO
           END DO

           ! Dimensions
           IF (NOMULT.GT.NDEFOR) STOP 'MASPAR: INCREASE NDEFOR PARAMETER!'

       END IF

       !  Start calculations

       CALL CPUTIM('MASPAR',1)

       !-------------------------------------------------------------------!
       !         ENERGY MOMENTS N=1..NDEMOM FOR ALL MASS COMPONENTS
       !-------------------------------------------------------------------!

       ! Calculate energy moments (EMOMET)
       CALL INVEMO(MREVER,ICHARG)

       !-------------------------------------------------------------------!
       !
       !  ATTENTION!!!
       !
       !     Factor (1/2) in the definitions of sigma^(k) comes from the
       !     difference in the way we calculate moments m^k  where summs
       !     run over the whole set of indices mu,nu and in sigma (where
       !     the sums run over the positive indices only).
       !
       !     This assumes the following time-reversal property of the
       !     collective variable operators q: |<0|q|-m-n>|=|<0|q|mn>|
       !
       !-------------------------------------------------------------------!

       !-------------------------------------------------------------------!
       !          I.  ATDHF (CRANKING APPROXIMATION ONLY)                  !
       !-------------------------------------------------------------------!

       ! Mass parameters B_{ij} given by Eq.(23), Libert et al. (1999)
       AUX1=0.0D0
       AUX2=0.0D0
       AUX3=0.0D0
       ! M-1**-1 * M-3 * M-1**-1 --> AUX1
       AUX1=EMOMET(:,:,1)
       AUX2=EMOMET(:,:,3)
       AUX3=0.0D0
       CALL INVMUL(AUX1,AUX2,AUX3,NOMULT,NDEFOR)
       ! Pack results in the matrix ATDMAS(:,ICHARG)
       FACTOR=0.25
       AUX1 = FACTOR*AUX1
       CALL S2VCPY(AUX1,NOMULT,NDEFOR,ATDMAS(1,ICHARG),NOCOMP,NDCOMP)

       ! Zero-point energy given by E0_ATD = Tr[ M2 M3^(-1) ] (Eq. (32), Libert et al., (1999)
       AUX1=0.0D0
       AUX2=EMOMET(:,:,2)
       AUX3=EMOMET(:,:,3)
       ! Invert M3 moment
       CALL INVERT(AUX3,AUX1,NOMULT,NDEFOR)
       ! Multiply M2 by 1/M3
       AUX1=MATMUL(AUX2,AUX3)
       ! Get trace, save
       E0_ATD(ICHARG)=0.5*TRACEM(AUX1,NOMULT,NDEFOR)

       !-------------------------------------------------------------------!
       !     C R A N K I N G
       !-------------------------------------------------------------------!

       ! M1**-1 * M3 * M1**-1 --> AUX1
       AUX1=EMOMET(:,:,1)
       AUX2=EMOMET(:,:,3)
       AUX3=0.0D0

       CALL INVMUL(AUX1,AUX2,AUX3,NOMULT,NDEFOR)

       FACTOR=0.25
       AUX1=FACTOR*AUX1

       ! Pack results into the matrix CRAMAS(:,ICHARG)
       CALL S2VCPY(AUX1,NOMULT,NDEFOR,CRAMAS(1,ICHARG),NOCOMP,NDCOMP)

       !-------------------------------------------------------------------!
       !          II.  GCM                                                 !
       !-------------------------------------------------------------------!

       ! Mass parameters B_GCM = 2 S(2) S(1)^-1 S(2), Eq.(61), Baran et al, (2011)

       ! S(2) = 1/4 * 1/2 M1^(-1) * M2 * M1^(-1) --> AUX2
       AUX1=EMOMET(:,:,2)
       AUX2=EMOMET(:,:,1)
       AUX3=0.0D0
       CALL INVMUL(AUX2,AUX1,AUX3,NOMULT,NDEFOR)
       FACTOR=0.125
       AUX2=FACTOR*AUX2
       ! Save metric S(2)
       COVMET(:,:,ICHARG)=AUX2

       ! S(1)^-1 = [ 1/4*1/2 M1^(-1) * M1 * M1^(-1) ]^-1 = 4 M1
       AUX1=EMOMET(:,:,1)
       FACTOR=8.0
       AUX1=FACTOR*AUX1

       ! B_ij = 2 * S(2) * S(1)^(-1) * S(2)
       AUX2=MATMUL(AUX2,MATMUL(AUX1,AUX2))
       FACTOR=2.0
       AUX2=FACTOR*AUX2

       ! Save partial, covariant GCM mass
       CALL S2VCPY(AUX2,NOMULT,NDEFOR,GCMMAS(1,ICHARG),NOCOMP,NDCOMP)

       ! E0_GCM = 1/4 Tr( Sig2^(-1) Sig1 ) (SEE: Pomorski NP, 1989) (16)
       ! GET METRIC (SIG2) AND INVERT IT
       AUX2=COVMET(:,:,ICHARG)
       CALL INVERT(AUX2,AUX3,NOMULT,NDEFOR)

       ! Make S1=1/4 M1^-1
       AUX1=EMOMET(:,:,1)
       CALL INVERT(AUX1,AUX3,NOMULT,NDEFOR)
       FACTOR=0.125
       AUX1=FACTOR*AUX1

       ! Make SIG2^-1*SIG1
       AUX3=MATMUL(AUX2,AUX1)

       ! Get trace of 1/4*(SIG2^-1 SIG1)
       E0_GCM(ICHARG)=0.25*TRACEM(AUX3,NOMULT,NDEFOR)

       !  .. TEST,  IBID. (17) ..
       !
       !   E0T_16=E0_GCM(ICHARG)
       !   AUX1=COVMET(:,:,ICHARG)
       !   CALL V2SCPY(GCMMAS(1,ICHARG),NOCOMP,NDCOMP,AUX2,NOMULT,NDEFOR)
       !   CALL INVERT(AUX2,AUX3,NOMULT,NDEFOR)
       !   AUX3=MATMUL(AUX1,AUX2)
       !   E0T_17=0.5*TRACEM(AUX3,NOMULT,NDEFOR)
       !   WRITE(NFIPRI,*) E0T_16,E0T_17

       !-------------------------------------------------------------------!
       !     PRINTING MASS TENSOR
       !-------------------------------------------------------------------!

       ! Write partial mass and zero-point energy for neutrons/protons
       CALL MASPRI(ICHARG)

       ! Calculate and write total mass and zero point energy for the
       ! nucleus
       IF (ICHARG.EQ.1) THEN
           CALL TOTMAS
           CALL MASPRI(2)
       END IF

       CALL CPUTIM('MASPAR',0)

       RETURN
     END SUBROUTINE MASPAR

     !--------------------------------------------------------------------!
     !   This routine calculates a composite (total) contravariant mass   !
     !   tensor for (n + p)-system and packs tensor components into the   !
     !   matrix GCMMAS(:,2) in the case of GCM and adds masses in other   !
     !   cases (ATDHF and cranking)                                       !
     !--------------------------------------------------------------------!
     SUBROUTINE TOTMAS

       REAL, DIMENSION (NDEFOR,NDEFOR) :: AUX1,AUX2,AUX3

       ! Initialization of auxiliary arrays

       AUX1 = 0.0D0
       AUX2 = 0.0D0
       AUX3 = 0.0D0

       !------------------------------------------------------------------!
       !                           Masses                                 !
       !------------------------------------------------------------------!

       ATDMAS(:,2)=ATDMAS(:,0)+ATDMAS(:,1)
       CRAMAS(:,2)=CRAMAS(:,0)+CRAMAS(:,1)

       !  GCM case
       !
       !   BT = (GP + GN) * [GP*BP^(-1)*GP + GN*BN^(-1)*GN]^(-1) * (GP + GN)
       !
       !   G = Matrix of covariant metric tensor
       !   N = neutrons, P = protons, T = total

       ! Total metric (cov)
       COVMET(:,:,2)=COVMET(:,:,0)+COVMET(:,:,1)

       ! Get n-mass and p-mass tensors, respectively
       CALL V2SCPY(GCMMAS(1,0),NOCOMP,NDCOMP,AUX1,NOMULT,NDEFOR)
       CALL V2SCPY(GCMMAS(1,1),NOCOMP,NDCOMP,AUX2,NOMULT,NDEFOR)

       ! Invert masses for each isospin (-> contravariant (M^-1)^KL)
       CALL INVERT(AUX1,AUX3,NOMULT,NDEFOR)
       AUX3 = 0.0D0

       CALL INVERT(AUX2,AUX3,NOMULT,NDEFOR)
       AUX3 = 0.0D0

       ! Make them covariant (-->(M^-1)^KL)
       AUX3=COVMET(:,:,0)
       AUX1=MATMUL(AUX3,MATMUL(AUX1,AUX3))

       AUX3=COVMET(:,:,1)
       AUX2=MATMUL(AUX3,MATMUL(AUX2,AUX3))

       ! Add covariant components (M^-1 = M^-1_n + M^-1_n)
       AUX3=AUX1+AUX2

       ! Invert
       CALL INVERT(AUX3,AUX1,NOMULT,NDEFOR)

       ! Get total metric
       AUX2=COVMET(:,:,2)

       ! Make B_GCM
       AUX2=MATMUL(AUX2,MATMUL(AUX3,AUX2))

       ! Pack total GCM mass --> GCMMAS(:,2)
       CALL S2VCPY(AUX2,NOMULT,NDEFOR,GCMMAS(1,2),NOCOMP,NDCOMP)

       !------------------------------------------------------------------!
       !                     Zero-point energies                          !
       !------------------------------------------------------------------!

       !  GCM: E0 = 1/2 Tr[ (gamma B)^(-1) ] (see Pomorski NP, 1989)

       ! Get total metric (SIG2)
       AUX2=COVMET(:,:,2)

       ! Get total mass
       AUX1 = 0.0D0
       CALL V2SCPY(GCMMAS(1,2),NOCOMP,NDCOMP,AUX1,NOMULT,NDEFOR)

       ! Invert
       CALL INVERT(AUX1,AUX3,NOMULT,NDEFOR)

       ! Multiply
       AUX3 = 0.0D0
       AUX3=MATMUL(AUX1,AUX2)

       ! Trace/2
       E0_GCM(2)=0.5*TRACEM(AUX3,NOMULT,NDEFOR)

       !  ATDHF: add both ICHARG components
       E0_ATD(2)=E0_ATD(0)+E0_ATD(1)

       RETURN
     END SUBROUTINE TOTMAS

     !--------------------------------------------------------------------!
     !  THis routine adds n and p and/or writes mass tensor components    !
     !                                                                    !
     !  Meaning of some parameters                                        !
     !    ICHARG - Standard 0, 1 = N, P; WRITE                            !
     !                         2 = add, form total components and write   !
     !    NOMULT - Number of constraints (LAMBDA>=2)                      !
     !    NOCOMP - Total number of different components of mass tensors   !
     !                                                                    !
     !--------------------------------------------------------------------!
     SUBROUTINE MASPRI(ICHARG)

       COMMON /CFIPRI/ NFIPRI

       CHARACTER(LEN=8) NAMEPN
       CHARACTER(LEN=9) MEVBCH
       CHARACTER(LEN=2) DIVBY2,EMPTCH,OHAFCH

       !  Particle type
       IF (ICHARG.EQ.0) NAMEPN='NEUTRONS'
       IF (ICHARG.EQ.1) NAMEPN=' PROTONS'
       IF (ICHARG.EQ.2) NAMEPN='   TOTAL'

       !  Mass parameter units
       OHAFCH='/2'
       EMPTCH='  '
       MEVBCH='1/MeV b**'

       WRITE(NFIPRI,'(79(''*''),/,''*'',77X,''*'')')
       WRITE(NFIPRI,'(''*'',29X,''COLLECTIVE INERTIA'',30X,''*'')')
       WRITE(NFIPRI,'(''*'',77X,''*'')')
       WRITE(NFIPRI,'(''*  MASS TENSOR B_{ij}/HbarC**2 WITH i=(L,M) AND j=(LP,MP)'',8X,A8,5X''*'')') NAMEPN
       WRITE(NFIPRI,'(''*'',77X,''*'')')
       WRITE(NFIPRI,'(''*    L   M    LP  MP       ATDHF       CRANKING      GCM-GOA       UNITS      *'')')

       DO ICOMPS=1,NOCOMP

          IUNITS=INDMAS(ICOMPS,1)+INDMAS(ICOMPS,3)
          DIVBY2=OHAFCH

          IF (MOD(IUNITS,2).EQ.0) THEN
              DIVBY2=EMPTCH
              IUNITS=IUNITS/2
          ELSE
              DIVBY2=OHAFCH
          END IF

          WRITE(NFIPRI,'(''*'',1X,2I4,2X,2I4,3E14.4,2X,A9,I1,A2,2X,''*'')') &
                INDMAS(ICOMPS,1),INDMAS(ICOMPS,2), &
                INDMAS(ICOMPS,3),INDMAS(ICOMPS,4), &
                ATDMAS(ICOMPS,ICHARG), &
                CRAMAS(ICOMPS,ICHARG), &
                GCMMAS(ICOMPS,ICHARG),MEVBCH,IUNITS,DIVBY2

       END DO

       ! Zero point energy
       WRITE(NFIPRI,'(''*'',77X,''*'')')
       WRITE(NFIPRI,'(''*  ZERO-POINT ENERGY CORRECTIONS - ATDHF:'',F9.4,'' MeV'',24X,''*'')') E0_ATD(ICHARG)
       WRITE(NFIPRI,'(''*'',36X,''GCM:'',F9.4,'' MeV'',24X,''*'')') E0_GCM(ICHARG)
       WRITE(NFIPRI,'(''*'',77X,''*'',/,79(''*''),/)')

       RETURN
     END SUBROUTINE MASPRI

     !--------------------------------------------------------------------!
     !                                                                    !
     !   This subroutine calculates inverse energy moments EMOMET(:.:.K)  !
     !   and pairing field corrections f(k) (see e.g. Gozdz...)           !
     !                                                                    !
     !   Meaning of some parameters (see also MASPAR routine):            !
     !     INDMAS(NOCOMP,4) - Mass tensor multi-indices                   !
     !     NOCOMP           - Actual number of components                 !
     !     NDEFOR           - Maximal number of deformations (multipoles) !
     !     FOPQPA(:,:)      - Matrix elements of corresponding  multipole !
     !     FOPQPB(:,:)        Operators calculated in subroutine INTMUL() !
     !     EMOMET(:,:,:)    - Inverse energy moments                      !
     !     AUX(NDEMOM)      - Auxiliary vector                            !
     !                                                                    !
     !--------------------------------------------------------------------!
     SUBROUTINE INVEMO(MREVER,ICHARG)

       DIMENSION AUXILI(NDEMOM)

       COMPLEX :: C_ZERO
       COMPLEX, ALLOCATABLE :: SPELEM(:,:), FOPQPA(:,:), FOPQPB(:,:)
       COMPLEX, ALLOCATABLE :: FA11PL(:,:), FA11MI(:,:), FB11PL(:,:), FB11MI(:,:)

       ALLOCATABLE OPIMRE(:,:)

       COMMON &
              /QUAPAR/ ENQUAP(1:NDSTAT,0:NDREVE), &
                       V2QUAP(1:NDSTAT,0:NDREVE)
       COMMON &
              /COEMUL/ COMULT(0:NDMULT,0:NDOSCI,0:NDOSCI,1:NDKART)
       COMMON &
              /T_PHAS/ IPHAPP(0:NDYMAX,0:NDYMAX,0:NDKART), &
                       IPHAPM(0:NDYMAX,0:NDYMAX,0:NDKART), &
                       IPHAMP(0:NDYMAX,0:NDYMAX,0:NDKART), &
                       IPHAMM(0:NDYMAX,0:NDYMAX,0:NDKART)
       COMMON &
              /DIMENS/ LDBASE
       COMMON &
              /QPNUMS/ NUMBQP(0:NDREVE,0:NDISOS)
       COMMON &
              /CAOCNP/ V2OCNP(1:2*NDBASE,0:NDISOS)
!@@@ TEMP - TEMP - TEMP - TEMP - TEMP - TEMP - TEMP - TEMP - TEMP
       COMMON &
              /FER_IS/ FERMFN(1:NDBASE,0:NDISOS)
       COMMON &
              /GCTEMP/ TEMP_T
       COMMON &
              /T_FLAG/ IFTEMP
!@@@ TEMP - TEMP - TEMP - TEMP - TEMP - TEMP - TEMP - TEMP - TEMP

       IF(.NOT.ALLOCATED(OPIMRE)) THEN
          ALLOCATE(OPIMRE(1:NDBASE,1:NDBASE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('OPIMRE','INVEMO')
          OPIMRE=0.0D0
       END IF

       IF(.NOT.ALLOCATED(SPELEM)) THEN
          ALLOCATE (SPELEM(1:NDBASE,1:NDBASE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('SPELEM','INVEMO')
          SPELEM=CMPLX(0.0D0, 0.0D0)
       END IF

       IF(.NOT.ALLOCATED(FOPQPA)) THEN
          ALLOCATE (FOPQPA(1:NDBASE,1:NDBASE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('FOPQPA','INVEMO')
          FOPQPA=CMPLX(0.0D0, 0.0D0)
       END IF
       IF(.NOT.ALLOCATED(FOPQPB)) THEN
          ALLOCATE (FOPQPB(1:NDBASE,1:NDBASE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('FOPQPB','INVEMO')
          FOPQPB=CMPLX(0.0D0, 0.0D0)
       END IF

       IF (IFTEMP.EQ.1) THEN
          ALLOCATE(FA11PL(1:NDBASE,1:NDBASE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('FA11PL','INVEMO')
          FA11PL(:,:)=CMPLX(0.0D0, 0.0D0)
          ALLOCATE(FA11MI(1:NDBASE,1:NDBASE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('FA11MI','INVEMO')
          FA11MI=CMPLX(0.0D0, 0.0D0)
          ALLOCATE(FB11PL(1:NDBASE,1:NDBASE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('FB11PL','INVEMO')
          FB11PL=CMPLX(0.0D0, 0.0D0)
          ALLOCATE(FB11MI(1:NDBASE,1:NDBASE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('FB11MI','INVEMO')
          FB11MI=CMPLX(0.0D0, 0.0D0)
       END IF

       C_ZERO=CMPLX(0.0D0,0.0D0)

       ! Number of qp-states
       NUM_QP=NUMBQP(0,ICHARG)

       ! Old indices
       LAMAOL=-1
       MIUAOL=-1
       LAMBOL=-1
       MIUBOL=-1

       ! Flag controlling the symmetry properties of the multipoole
       ! moment operators
       ISIPAR=1
       ITIPAR=1
       IHERMI=1
       IMAREA=1

       ICOMPS=0

       ! Loop over the collective variables a and b, referenced by
       ! indexes ICOLLA and ICOLLB

       DO ICOLLA=1,NOMULT
          DO ICOLLB=ICOLLA,NOMULT

             ! CASE 1: collective variables are only multipole moments
             IF (IIPAIR.NE.1) THEN

                 ICOMPS=ICOMPS+1

                 LAMBDA=INDMAS(ICOMPS,1)
                 MIUA  =INDMAS(ICOMPS,2)

                 LAMBDB=INDMAS(ICOMPS,3)
                 MIUB  =INDMAS(ICOMPS,4)

                 ! Are new Q(A) and old Q(A) different?
                 IF (LAMAOL.NE.LAMBDA.OR.MIUAOL.NE.MIUA) THEN
                    !
                    ! If Q(A) and old Q(B) are the same, copy old B to A
                    IF (LAMBDA.EQ.LAMBOL.AND.MIUA.EQ.MIUBOL) THEN
                        FOPQPA=FOPQPB
                        FA11PL=FB11PL
                        FA11MI=FB11MI
                    ELSE

                        ! Make new Q(LAMBDA,MIUA)
                        CALL INTMUL(IPHAPP(0,0,0),OPIMRE,COMULT,LAMBDA,MIUA,IMAREA)

                        SPELEM=CMPLX(OPIMRE,0.D0)

                        CALL F2OPER(SPELEM,FOPQPA,ICHARG,MREVER)

                        IF (IFTEMP.EQ.1) THEN
                        	CALL F11OPE(SPELEM,FA11PL,FA11MI,ICHARG,MREVER)
                        END IF

                    END IF
                 END IF

                 ! Are new Q(B) and old Q(B) different?
                 IF (LAMBOL.NE.LAMBDB.OR.MIUBOL.NE.MIUB) THEN

                    ! Are Q(A) and Q(B) the same?
                    IF (LAMBDB.EQ.LAMBDA.AND.MIUB.EQ.MIUA) THEN
                        FOPQPB=FOPQPA
                        FB11PL=FA11PL
                        FB11MI=FA11MI
                    ELSE

                        ! Make new Q(LAMBDB,MIUB)
                        CALL INTMUL(IPHAPP(0,0,0),OPIMRE,COMULT,LAMBDB,MIUB,IMAREA)

                        SPELEM=CMPLX(OPIMRE,0.D0)

                        CALL F2OPER(SPELEM,FOPQPB,ICHARG,MREVER)

                        IF (IFTEMP.EQ.1) THEN
                        	CALL F11OPE(SPELEM,FB11PL,FB11MI,ICHARG,MREVER)
                        END IF

                    END IF
                 END IF

             ! CASE 2: Collective variables include multipole moments and lambda_2
             ELSE

                 ! Subcase 1: part of the inertia tensor involving only multipole moments
                 IF (ICOLLA.NE.NOMULT)THEN
                     ICOMPS=ICOMPS+1
                     LAMBDA=INDMAS(ICOMPS,1)
                     MIUA  =INDMAS(ICOMPS,2)
                     LAMBDB=INDMAS(ICOMPS,3)
                     MIUB  =INDMAS(ICOMPS,4)
                     !    ______ Q(A) - INDEX ______
                     CALL INTMUL(IPHAPP(0,0,0),OPIMRE,COMULT,LAMBDA,MIUA,IMAREA)
                     SPELEM=CMPLX(OPIMRE,0.D0)
                     CALL F2OPER(SPELEM,FOPQPA,ICHARG,MREVER)
                     !   ______ Q(B) - INDEX ______
                     CALL INTMUL(IPHAPP(0,0,0),OPIMRE,COMULT,LAMBDB,MIUB,IMAREA)
                     SPELEM=CMPLX(OPIMRE,0.D0)
                     CALL F2OPER(SPELEM,FOPQPB,ICHARG,MREVER)

                 ! Subcase 2: part of the inertia tensor where variable A is a multipole moment
                 !	      and variable B is a lambda_2 parameter
                 ELSE IF(ICOLLA.EQ.NOMULT.AND.ICOLLB.NE.ICOLLA) THEN

                     ICOMPS=ICOMPS+1
                     LAMBDA=INDMAS(ICOMPS,1)
                     MIUA  =INDMAS(ICOMPS,2)
                     !    ______ Q(A) - INDEX ______
                     CALL INTMUL(IPHAPP(0,0,0),OPIMRE,COMULT,LAMBDA,MIUA,IMAREA)
                     SPELEM=CMPLX(OPIMRE,0.D0)
                     CALL F2OPER(SPELEM,FOPQPA,ICHARG,MREVER)
                     !    ______ Q(B) - INDEX ______
                     SPELEM=C_ZERO
                     DO ISTATE=1,LDBASE

                        IF (V2OCNP(ISTATE,ICHARG).GE.0.0D0) THEN
                            V_ACTU=SQRT(V2OCNP(ISTATE,ICHARG))
                        ELSE
                            V_ACTU=0.0D0
                        END IF

                        IF (V2OCNP(ISTATE,ICHARG).LE.1.0D0) THEN
                            U_ACTU=SQRT(1.0D0-V2OCNP(ISTATE,ICHARG))
                        ELSE
                            U_ACTU=0.0D0
                            V_ACTU=1.0D0
                        END IF

                        S_U2V2 = 2.0D0 * (1.0D0 -2.0D0* V_ACTU**2)
                        SPELEM(ISTATE,ISTATE)=CMPLX(S_U2V2,0.0D0)
                     END DO

                    CALL F2OPER(SPELEM,FOPQPB,ICHARG,MREVER)

                 ! Subcase 2: part of the inertia tensor where both collective variables are
                 !	      lambda_2 parameters
                 ELSE IF(ICOLLA.EQ.NOMULT.AND.ICOLLB.EQ.ICOLLA) THEN

                    SPELEM=C_ZERO
                    DO ISTATE=1,LDBASE

                       IF (V2OCNP(ISTATE,ICHARG).GE.0.0D0) THEN
                           V_ACTU=SQRT(V2OCNP(ISTATE,ICHARG))
                       ELSE
                           V_ACTU=0.0D0
                       END IF

                       IF (V2OCNP(ISTATE,ICHARG).LE.1.0D0) THEN
                           U_ACTU=SQRT(1.0D0-V2OCNP(ISTATE,ICHARG))
                       ELSE
                           U_ACTU=0.0D0
                           V_ACTU=1.0D0
                       END IF

                       S_U2V2 = 2.0D0 * (1.0D0 - 2.0D0*V_ACTU**2)
                       SPELEM(ISTATE,ISTATE)=CMPLX(S_U2V2,0.0D0)
                    END DO
                    CALL F2OPER(SPELEM,FOPQPA,ICHARG,MREVER)
                    ! In the q.p. basis, the matrix of variable B is the same as that of variable !
                    FOPQPB=FOPQPA

                END IF

             END IF

             !  Computing the inverse moments M_{-k}^{ab}
             !  Reference:
             !       Eq. (24), Libert et al., Phy. Rev. C 60, 054301 (1999)
             !       Eq. (3.39), Gozdz et al., Nucl. Phys. A 442, 26 (1985)

             AUXILI(:)=0.0D0

             ! Sums run over time-even states (ibra>0; -iket>0)
             DO IQP_MU=1,NUM_QP
                EQP_MU=ENQUAP(IQP_MU,0)
                DO IQP_NU=1,NUM_QP
                   EQP_NU=ENQUAP(IQP_NU,0)

                   SUMMEQ=EQP_MU+EQP_NU
                   ELPROD=REAL(FOPQPA(IQP_MU,IQP_NU)*CONJG(FOPQPB(IQP_MU,IQP_NU)))

                   ! Finite-temperature correction
                   IF (IFTEMP.EQ.1) THEN

                       T1TERM=1.0D0 + FERMFN(IQP_MU,0) + FERMFN(IQP_NU,0)

                       TEMP_K=FERMFN(IQP_MU,0)
                       TEMP_L=FERMFN(IQP_NU,0)

                       E11PLU=REAL(FA11PL(IQP_MU,IQP_NU)*CONJG(FB11PL(IQP_MU,IQP_NU)))
                       E11MIN=REAL(FA11MI(IQP_MU,IQP_NU)*CONJG(FB11MI(IQP_MU,IQP_NU)))

 	               IF (IQP_NU.NE.IQP_MU) THEN
                           DIFFEQ=EQP_MU-EQP_NU
                           T2TERM=TEMP_K-TEMP_L * 0.5D0*(E11PLU + E11MIN)/DIFFEQ**K
                       ELSE
                           T2TERM=TEMP_K*(TEMP_K-1.0D0)/TEMP_T * 0.5D0*(E11PLU + E11MIN)
                       END IF

                   ELSE
                       T1TERM=1.0D0
                       T2TERM=0.0D0
                   END IF

                   ! Inverse energy moments ENMOMS(N), N = 1, .., NDEMOM
                   DO K=1,NDEMOM
                      AUXILI(K)=AUXILI(K)+ELPROD/SUMMEQ**K * T1TERM + T2TERM
                   END DO

                END DO  ! IQP_NU
             END DO ! IQP_MU

             ! Save inverse energy moments
             FACTOR=2.0D0
             DO K=1,NDEMOM
                EMOMET(ICOLLA,ICOLLB,K)=FACTOR*AUXILI(K)
                EMOMET(ICOLLB,ICOLLA,K)=EMOMET(ICOLLA,ICOLLB,K)
             END DO

             ! Replace old indices
             LAMAOL=LAMBDA
             MIUAOL=MIUA

             LAMBOL=LAMBDB
             MIUBOL=MIUB

          END DO ! ICOLLB
       END DO ! ICOLLA

       DEALLOCATE(SPELEM,OPIMRE,FOPQPA,FOPQPB)
       IF (IFTEMP.EQ.1) THEN
       	   DEALLOCATE(FA11PL,FA11MI,FB11PL,FB11MI)
       END IF

       RETURN
     END SUBROUTINE INVEMO

     !--------------------------------------------------------------------!
     !                                                                    !
     ! This routine calculates matrix elements <0| \beta_i \beta_j F |0>  !
     ! where |0> is the HFB vacuum (possibly under constraints), \beta_i  !
     ! and \beta_j are qp-annihilation operators and F is a one-body      !
     ! hermitian operator:                                                !
     !                        F = sum_{kl} F_{kl} a^{+}_{k}a_{l}          !
     !                                                                    !
     ! with a^{+}_{k} and a_{l} single-particle creation and anihilation  !
     ! operators, respectively.                                           !
     !                                                                    !
     !--------------------------------------------------------------------!
     SUBROUTINE F2OPER(SPOPER,F2ELEM,ICHARG,MREVER)

       USE ALLQUA

       COMPLEX   C_ZERO,C_UNIT,C_MINU
       COMPLEX   F2ELEM,SPOPER,DUMMY1

       DIMENSION SPOPER(1:NDBASE,1:NDBASE), &
                 F2ELEM(1:NDBASE,1:NDBASE), &
                 DUMMY1(1:NDBASE,1:NDBASE)

       COMMON /QPNUMS/ NUMBQP(0:NDREVE,0:NDISOS)
       COMMON /DIMENS/ LDBASE

       C_ZERO=CMPLX( 0.0D0,0.0D0)
       C_UNIT=CMPLX(+1.0D0,0.0D0)
       C_MINU=CMPLX(-1.0D0,0.0D0)

       IF (.NOT.ALLOCATED(AWAQUA)) THEN
          ALLOCATE (AWAQUA(1:NDBASE,1:NDSTAT,0:NDREVE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('AWAQUA','F2OPER')
       END IF

       IF (.NOT.ALLOCATED(BWAQUA)) THEN
          ALLOCATE (BWAQUA(1:NDBASE,1:NDSTAT,0:NDREVE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('BWAQUA','F2OPER')
       END IF

       F2ELEM = C_ZERO
       DUMMY1 = C_ZERO

       ! In HFODD simplex basis, the operator F (SPOPER) has a block
       ! structure
       !               F = [  F_{+}   0   ]
       !	             [    0   F_{-} ]
       ! Simplifications:
       !   - if time-reversal symmetry is conserved, F_{-} = F_{+}^{*}.
       !   - if, in addition, the operator F is real, then F_{-} = F_{+}
       !
       ! The component F_{20} of the operator F in the q.p. basis reads
       ! (most generally).
       !
       !      F_{20} = U_{-}^{+} F_{-} V_{-}^{*}
       !             - V_{+}^{+} F_{+}^{*} U_{+}^{*}
       ! Below we compute this representation for operators SPOPER that
       ! are the multipole moment operators (they are real in the simplex
       ! basis), and in the case of time-reversal symmetry.

       !  Second term: V_{+}^{+} F_{+}^{*} U_{+}^{*}
       !             = B_{+}^{T} F_{+}     A_{+} = B_{+}^{T} F_{+}^{T}A_{+}

       CALL ZGEMM('T','T',NUMBQP(0,ICHARG),LDBASE,LDBASE, &
                             C_UNIT,BWAQUA(1,1,0),NDBASE, &
                                           SPOPER,NDBASE, &
                                    C_ZERO,DUMMY1,NDBASE)

       CALL ZGEMM('N','N',NUMBQP(0,ICHARG),NUMBQP(0,ICHARG),LDBASE, &
                                              C_UNIT,DUMMY1,NDBASE, &
                                              AWAQUA(1,1,0),NDBASE, &
                                              C_ZERO,F2ELEM,NDBASE)

       ! First term:  U_{-}^{+} F_{-} V_{-}^{*}
       !            = A_{-}^{T} F_{-} B_{-} = A_{-}^{T} F_{-}^{T} B_{-}

       CALL ZGEMM('T','T',NUMBQP(1,ICHARG),LDBASE,LDBASE, &
                             C_UNIT,AWAQUA(1,1,1),NDBASE, &
                                           SPOPER,NDBASE, &
                                    C_ZERO,DUMMY1,NDBASE)

       CALL ZGEMM('N','N',NUMBQP(1,ICHARG),NUMBQP(1,ICHARG),LDBASE, &
                                              C_UNIT,DUMMY1,NDBASE, &
                                              BWAQUA(1,1,1),NDBASE, &
                                              C_MINU,F2ELEM,NDBASE)

       RETURN
     END SUBROUTINE F2OPER

     !--------------------------------------------------------------------!
     !                                                                    !
     ! This routine calculates matrix elements <0| \beta_i \beta_j F |0>  !
     ! where |0> is the HFB vacuum (possibly under constraints), \beta_i  !
     ! and \beta_j are qp-annihilation operators and F is a one-body      !
     ! hermitian operator:                                                !
     !                        F = sum_{kl} F_{kl} a^{+}_{k}a_{l}          !
     !                                                                    !
     ! with a^{+}_{k} and a_{l} single-particle creation and anihilation  !
     ! operators, respectively.                                           !
     !                                                                    !
     !--------------------------------------------------------------------!
     SUBROUTINE F11OPE(SPOPER,FN11PL,FN11MI,ICHARG,MREVER)

       USE ALLQUA

       COMPLEX :: C_ZERO,C_UNIT,C_MINU
       COMPLEX :: FN11PL,FN11MI,SPOPER,CMPLXN

       DIMENSION SPOPER(1:NDBASE,1:NDBASE), &
                 FN11PL(1:NDBASE,1:NDBASE), &
                 FN11MI(1:NDBASE,1:NDBASE)

       ALLOCATABLE CMPLXN(:,:)

       COMMON /QPNUMS/ NUMBQP(0:NDREVE,0:NDISOS)
       COMMON /DIMENS/ LDBASE

       C_ZERO=CMPLX( 0.0D0,0.0D0)
       C_UNIT=CMPLX(+1.0D0,0.0D0)
       C_MINU=CMPLX(-1.0D0,0.0D0)

       IF (.NOT.ALLOCATED(AWAQUA)) THEN
          ALLOCATE (AWAQUA(1:NDBASE,1:NDSTAT,0:NDREVE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('AWAQUA','F11OPE')
       END IF

       IF (.NOT.ALLOCATED(BWAQUA)) THEN
          ALLOCATE (BWAQUA(1:NDBASE,1:NDSTAT,0:NDREVE),STAT=IALLOC)
          IF (IALLOC.NE.0) CALL NOALLO('BWAQUA','F11OPE')
       END IF

       ALLOCATE (CMPLXN(1:NDBASE,1:NDSTAT),STAT=IALLOC)
       IF (IALLOC.NE.0) CALL NOALLO('CMPLXN','F11OPE')

       CMPLXN(:,:) = C_ZERO

       !  Computing \tilde{F}^{11}
       !
       !  Second term: V_{+}^{+} F_{+}^{*} V_{+}
       !             = B_{+}^{T}(B_{+}^{T} F_{+}^{T})^{+}
       !
       CALL ZGEMM('T','T',NUMBQP(0,0),LDBASE,LDBASE, &
                        C_UNIT,BWAQUA(1,1,0),NDBASE, &
                                      SPOPER,NDBASE, &
                               C_ZERO,CMPLXN,NDSTAT)

       CALL ZGEMM('T','C',NUMBQP(0,0),NUMBQP(0,0),LDBASE, &
                             C_UNIT,BWAQUA(1,1,0),NDBASE, &
                                           CMPLXN,NDSTAT, &
                                    C_ZERO,FN11PL,NDSTAT)

       !  First term:  U_{-}^{+} F_{-} U_{-}
       !             = A_{-}^{T}(A_{-}^{T} F_{-}^{+})^{+}
       !
       CALL ZGEMM('T','C',NUMBQP(1,0),LDBASE,LDBASE, &
                        C_UNIT,AWAQUA(1,1,1),NDBASE, &
                                      SPOPER,NDBASE, &
                               C_ZERO,CMPLXN,NDSTAT)

       CALL ZGEMM('T','C',NUMBQP(1,0),NUMBQP(1,0),LDBASE, &
                             C_UNIT,AWAQUA(1,1,1),NDBASE, &
                                           CMPLXN,NDSTAT, &
                                    C_MINU,FN11PL,NDSTAT)



       CALL ZGEMM('T','T',NUMBQP(1,0),LDBASE,LDBASE, &
                        C_UNIT,BWAQUA(1,1,1),NDBASE, &
                                      SPOPER,NDBASE, &
                               C_ZERO,CMPLXN,NDSTAT)

       CALL ZGEMM('T','C',NUMBQP(1,0),NUMBQP(1,0),LDBASE, &
                             C_UNIT,BWAQUA(1,1,1),NDBASE, &
                                           CMPLXN,NDSTAT, &
                                    C_ZERO,FN11MI,NDSTAT)

       CALL ZGEMM('T','C',NUMBQP(0,0),LDBASE,LDBASE, &
                        C_UNIT,AWAQUA(1,1,0),NDBASE, &
                                      SPOPER,NDBASE, &
                               C_ZERO,CMPLXN,NDSTAT)

       CALL ZGEMM('T','C',NUMBQP(0,0),NUMBQP(0,0),LDBASE, &
                             C_UNIT,AWAQUA(1,1,0),NDBASE, &
                                           CMPLXN,NDSTAT, &
                                    C_MINU,FN11MI,NDSTAT)

       RETURN
     END SUBROUTINE F11OPE

     !--------------------------------------------------------------------!
     !  This routine copies a square symmetric matrix A into a vector V   !
     !         A  - Matrix to be copied into B                            !
     !         N  - Actual dimensions                                     !
     !         NA - Physical dimensions of A                              !
     !         V  - Resulting, lower triangular part of A                 !
     !         M  - Number of elements in B                               !
     !         MV - Physical dimension of V                               !
     !--------------------------------------------------------------------!
     SUBROUTINE S2VCPY(A,N,NA,V,M,MV)
       DIMENSION :: A(NA,NA), V(MV)
       IF (N.GT.NA.OR.M.GT.MV) STOP &
           'S2VCPY: N .GT. NA; PHYSICAL DIMENSIONS OF A OR V EXCEEDED'
       K=0
       DO I=1,N
          DO J=1,I
             K=K+1
             V(K)=A(I,J)
          END DO
       END DO
       RETURN
      END SUBROUTINE S2VCPY

     !--------------------------------------------------------------------!
     !  This routine copies the vector v(:) into a square matrix a(:,:)   !
     !    V(:)   - Vector to be copied                                    !
     !    A(:,:) - Resulting symmetric matrix                             !
     !    N      - Actual dimensions of A                                 !
     !    NA     - Physical dimension of A                                !
     !    M      - Actual dimension of V                                  !
     !    MV     - Physical dimension of V                                !
     !--------------------------------------------------------------------!
     SUBROUTINE V2SCPY(V,M,MV,A,N,NA)
       DIMENSION :: V(MV),A(NA,NA)
       IF (N.GT.NA.OR.M.GT.MV) STOP &
           'V2SCPY: N .GT. NA; PHYSICAL DIMENSIONS OF V OR A EXCEEDED'
       K=0
       DO I=1,N
          DO J=1,I
             K=K+1
             A(I,J)=V(K)
             A(J,I)=V(K)
          END DO
       END DO
       RETURN
     END SUBROUTINE V2SCPY

     !--------------------------------------------------------------------!
     !  This routine does special matrix multiplication with inversion    !
     !  of matrix A and saves results in A:  A <-- A**(-1) *B* A**(-1)    !
     !  XA is an auxiliary array of the same dimension as A.              !
     !--------------------------------------------------------------------!
     SUBROUTINE INVMUL(A,B,XA,N,NX)
       DIMENSION :: A(NX,NX),B(NX,NX),XA(NX,NX)
       IF (N.GT.NX) STOP &
           'INVMUL: N .GT. NX; PHYSICAL DIMENSIONS OF A, B EXCEEDED'
       CALL INVERT(A,XA,N,NX)
       A=MATMUL(A,MATMUL(B,A))
       RETURN
     END SUBROUTINE INVMUL

     !--------------------------------------------------------------------!
     !  This function calculates the trace of a matrix                    !
     !    A(:,:) - Input matrix                                           !
     !    N      - Actual dimension of A                                  !
     !    NX     - Physical dimensions of A                               !
     !--------------------------------------------------------------------!
     FUNCTION TRACEM(A,N,NX)
       DIMENSION :: A(NX,NX)
       IF (N.GT.NX) STOP 'TRACEM: DIMENSION N > NX!'
       TRACEM=0
       DO I=1,N
          TRACEM=TRACEM+A(I,I)
       END DO
       RETURN
     END FUNCTION TRACEM

     !--------------------------------------------------------------------!
     !  This subroutine calculates inverse of a matrix. A is the matrix   !
     !  to be inverted and is destroyed in the routine. Upon exit, matrix !
     !  A contains the inverse A^(-1).                                    !
     !                                                                    !
     !  Method ....: CHOLESKY LU decomposition + LU back substitution     !
     !  Routines ..: LUDCMP, LUBKSB  See: Num. Recipes., p 40)            !
     !--------------------------------------------------------------------!
     SUBROUTINE INVERT(A,Y,N,NP)
       PARAMETER (NMAX=500)
       DIMENSION :: A(NP,NP),Y(NP,NP),INDX(NMAX)
       IF (N.GT.NP.OR.N.GT.NMAX) &
           STOP 'INVERT: N TO LARGE. INCREASE NP OR NMAX IN INVERT.'
       ! Set up identity matrix
       Y(:,:)=0.0D0
       DO I=1,N
          Y(I,I)=1.0D0
       END DO

       ! Decompose the matrix just once
       CALL LUDCMP(A,N,NP,INDX,D)

       DO J=1,N
          ! Find inverse by columns
          CALL LUBKSB(A,N,NP,INDX,Y(1,J))
          ! Note that fortran stores two-dimensional matrices by column,
          ! so Y(1,J) is the address of the J-th column of Y.
       END DO

       ! Rewrite Y to A
       A=Y

       RETURN
     END SUBROUTINE INVERT

     SUBROUTINE LUDCMP(a,n,np,indx,d)
       PARAMETER (NMAX=500,TINY=1.0e-20)
       DIMENSION a(np,np),vv(NMAX),indx(n)
       INTEGER i,imax,j,k
       d=1.
       imax=1
       do i=1,n
          aamax=0.
          do j=1,n
             if (abs(a(i,j)).gt.aamax) aamax=abs(a(i,j))
          end do ! j
          if (aamax.eq.0.) write(*,*) 'singular matrix in ludcmp'
          vv(i)=1./aamax
       end do
       do j=1,n
          do i=1,j-1
             sum=a(i,j)
             do k=1,i-1
                sum=sum-a(i,k)*a(k,j)
             end do ! k
             a(i,j)=sum
          end do ! i
          aamax=0.
          do i=j,n
             sum=a(i,j)
             do k=1,j-1
                sum=sum-a(i,k)*a(k,j)
             end do ! k
             a(i,j)=sum
             dum=vv(i)*abs(sum)
             if (dum.ge.aamax) then
                imax=i
                aamax=dum
             end if
          end do ! i
          if (j.ne.imax) then
             do k=1,n
                dum=a(imax,k)
                a(imax,k)=a(j,k)
                a(j,k)=dum
             end do
             d=-d
             vv(imax)=vv(j)
          end if
          indx(j)=imax
          if(a(j,j).eq.0.) a(j,j)=TINY
          if(j.ne.n) then
             dum=1./a(j,j)
             do i=j+1,n
                a(i,j)=a(i,j)*dum
             end do
          end if
       end do ! j

       return
     END SUBROUTINE ludcmp

     SUBROUTINE lubksb(a,n,np,indx,b)
       INTEGER n,np,indx(n)
       DIMENSION a(np,np),b(n)
       INTEGER i,ii,j,ll
       ii=0
       do i=1,n
          ll=indx(i)
          sum=b(ll)
          b(ll)=b(i)
          if (ii.ne.0) then
             do j=ii,i-1
                sum=sum-a(i,j)*b(j)
             end do
          else if (sum.ne.0.) then
             ii=i
          end if
          b(i)=sum
       end do
       do i=n,1,-1
          sum=b(i)
          do j=i+1,n
             sum=sum-a(i,j)*b(j)
          end do
          b(i)=sum/a(i,i)
       end do

       return
     END SUBROUTINE lubksb

     !#######################################################################
     !    I N E R T I A   A N D  R O T A T I O N A L    C O R R E C T I O N S
     !#######################################################################

     !--------------------------------------------------------------------!
     !                                                                    !
     !  This subroutine calculates the approximate rotational correction  !
     !  to a HF energy:                                                   !
     !                      ECOR = -\sum_{k} J_k^{2} / (2*I_k)            !
     !                                                                    !
     !  Inertia parameters I_k, k=x,y,z are calculated in the cranking    !
     !  approximation. The J_k^{2} are the average values of the angular  !
     !  momentum operators in the BCS ground state. The routine is active !
     !  during the last iteration in HFOOD                                !
     !                                                                    !
     !  Usage:                                                            !
     !    In order to run the routine from HFODD one needs to add the     !
     !    following statement in the "neutron" and "proton" sections of   !
     !    the HFODD program (exact form):                                 !
     !                                                                    !
     !        CALL ROTCOR(TERMNT,ICHARG,MREVER,XMOMFC,YMOMFC,ZMOMFC,      !
     !   *                                     NXMAXX,NYMAXX,NZMAXX)      !
     !                                                                    !
     !  Dependencies                                                      !
     !    JMATEL - J_k matrices in simplex basis                          !
     !    ENMOMJ - Energy moments (0 and 1) of J_i                        !
     !    FLUCJ2 - J_K^2 fluctuations in the ground state                 !
     !                                                                    !
     !--------------------------------------------------------------------!
     SUBROUTINE ROTCOR(TERMNT,ICHARG,MREVER,XMOMFC,YMOMFC,ZMOMFC, &
                                            NXMAXT,NYMAXT,NZMAXT)

       COMMON &
              /MAXXYM/ NXMAXX,NYMAXX,NZMAXX
       COMMON &
              /CRAKER/ CRANKJ(0:1,1:3),FLUCTJ(0:1,1:3)
       COMMON &
              /CFIPRI/ NFIPRI

       DIMENSION :: CRANKT(3),FLUCTT(3),EROTAT(1:3),EMOMSJ(0:1)

       LOGICAL :: TERMNT

       DOUBLE PRECISION :: DEPSIL = 1.D-4

       ! Calculate only in case of the last HFB iteration
       IF (.NOT.TERMNT) RETURN

       CALL CPUTIM('ROTCOR',1)

       ! Dimensions from HFODD (used by JMATEL; COMMON /MAXXYM/)
       NXMAXX=NXMAXT
       NYMAXX=NYMAXT
       NZMAXX=NZMAXT

       IF (ICHARG.EQ.0) THEN
           CRANKJ(:,:)=0.0D0
           FLUCTJ(:,:)=0.0D0
       END IF

       DO IDIREC=1,3
          CALL ENMOMJ(EMOMSJ,IDIREC,MREVER,ICHARG)
          FLUCTJ(ICHARG,IDIREC)=0.5D0*EMOMSJ(0)
          CRANKJ(ICHARG,IDIREC)=      EMOMSJ(1)
       END DO

       ! Combine neutron and proton contributions and write data
       IF (ICHARG.EQ.1) THEN

           DO IDIREC=1,3
              CRANKT(IDIREC)=CRANKJ(0,IDIREC)+CRANKJ(1,IDIREC)
              FLUCTT(IDIREC)=FLUCTJ(0,IDIREC)+FLUCTJ(1,IDIREC)
              IF (CRANKT(IDIREC).LT.DEPSIL) THEN
                  EROTAT(IDIREC)=0.0D0
              ELSE
                  EROTAT(IDIREC)=-0.5D0*FLUCTT(IDIREC)/CRANKT(IDIREC)
              END IF
           END DO

           WRITE(NFIPRI,"(79(1H*),/,1H*,77x,1h*)")
           WRITE(NFIPRI,"(1H*,7X, &
              &  'MOMENTS OF INERTIA AND ROTATIONAL ENERGY CORRECTIONS', &
              &  ' (TOTAL)',10X,1H*)")
           WRITE(NFIPRI,"(1H*,77x,1h*)")

           WRITE(NFIPRI,"(1H*,5x,'RIGID:', &
              &  6x,'J_x=',F10.4,6x,'J_y=',F10.4,6x,'J_z=',F10.4,6x,1H*)") &
              &  XMOMFC,YMOMFC,ZMOMFC

           WRITE(NFIPRI, &
                "(1h*,77x,1H*,/,1H*,17x,54(1H-),6x,1H*,/,1H*,77x,1H*)")
           WRITE(NFIPRI,"(1H*,5x,'CRANK:', &
              &  6x,'J_x=',F10.4,6x,'J_y=',F10.4,6x,'J_z=',F10.4,6x,1H*)") &
              &  (CRANKT(K),K=1,3)

           WRITE(NFIPRI,"(1H*,77x,1H*)")
           WRITE(NFIPRI,"(1H*,5x,'FLUCT:', &
              &  6x,'jx2=',F10.4,6x,'jy2=',F10.4,6x,'jz2=',F10.4,6x,1h*)") &
              &  (FLUCTT(K),K=1,3)

           WRITE(NFIPRI,"(1H*,77x,1H*)")
           WRITE(NFIPRI,"(1H*,5x,'E_ROT:', &
              &  6x,'E_x=',F10.4,6x,'E_y=',F10.4,6x,'E_z=',F10.4,6X,1h*)") &
              &  (EROTAT(K),K=1,3)
           WRITE(NFIPRI,"(1h*,77x,1H*,/,79(1H*))")

       END IF

       CALL CPUTIM('ROTCOR',0)

       RETURN
     END SUBROUTINE ROTCOR

     !--------------------------------------------------------------------!
     !                                                                    !
     !  This subroutine returns 0-th and 1-st energy moments EMOMSJ(0)    !
     !  and EMOMSJ(1) of J_K, the angular momentum operators, for K=1=x,  !
     !  2=y, 3=z (parameter DIREC)                                        !
     !                                                                    !
     !  Definition:                                                       !
     !    EMOMSJ(N) = \sum_{kl} |J_{kl}|^2 * F_{kl} / (E_k+E_k)^N;        !
     !                where F_{kl} = (U_k*V_l - U_l*V_k)^22 and K,L>0,<0  !
     !                                                                    !
     !  W A R N I N G !!!                                                 !
     !  -----------------                                                 !
     !    The final results consist of contributions from both types of   !
     !    states namely T=+1 AND T=-1                                     !
     !                                                                    !
     !   Meaning of some parameters                                       !
     !     CELMAJ - Single particle matrix elements of J_I                !
     !     IDIREC - 1=x, 2=y, 3=z direction                               !
     !                                                                    !
     !--------------------------------------------------------------------!
     SUBROUTINE ENMOMJ(EMOMSJ,IDIREC,MREVER,ICHARG)

       COMPLEX :: CELMAJ,F2ELEM

       COMMON &
              /JOTJOT/ CELMAJ(1:NDBASE,1:NDBASE,0:NDREVE)
       COMMON &
              /SPNUMS/ NUMBSP(0:NDREVE,0:NDISOS)
       COMMON &
              /QPNUMS/ NUMBQP(0:NDREVE,0:NDISOS)
       COMMON &
              /PROCCU/ V_OCCU(1:2*NDSTAT)
       COMMON &
              /QUAPAR/ ENQUAP(1:NDSTAT,0:NDREVE), &
                       V2QUAP(1:NDSTAT,0:NDREVE)

       DIMENSION :: EMOMSJ(0:1)
       DIMENSION :: F2ELEM(1:NDBASE,1:NDBASE)

       ! Number of states
       LSTATE=NUMBQP(0,ICHARG)

       ! Matrix elements of J_k, k = IDIREC = 1(x), 2(y), 3(z)
       CALL JMATEL(MREVER,ICHARG,IDIREC)

       EMOMSJ(0)=0.0D0
       EMOMSJ(1)=0.0D0

       DO IREVER=0,NDREVE

          DO JSTATE=1,NUMBQP(IREVER,ICHARG)
             DO ISTATE=1,NUMBQP(IREVER,ICHARG)

                EDENOM=ENQUAP(ISTATE,0)+ENQUAP(JSTATE,0)
                ELMABS=ABS(CELMAJ(ISTATE,JSTATE,IREVER))**2

                EMOMSJ(0)=EMOMSJ(0)+ELMABS
                EMOMSJ(1)=EMOMSJ(1)+ELMABS/EDENOM

             END DO
          END DO

       END DO

       RETURN
     END SUBROUTINE ENMOMJ

     !--------------------------------------------------------------------!
     ! This subroutine calculates sp-matrix elements of J_k, k=x,y or z   !
     ! in the simplex basis ised in HFODD                                 !
     !--------------------------------------------------------------------!
     SUBROUTINE JMATEL(MREVER,ICHARG,INDEXJ)

       COMPLEX :: CELMAJ,CAUXIL
       COMPLEX :: RESULT,CRESJ2,CELMUL,CWAVPP,CWAVPM
       COMPLEX :: C_ZERO

       DIMENSION :: CAUXIL(1:NDBASE,1:NDBASE),RAUXIL(1:NDBASE,1:NDBASE)

       COMMON &
              /JOTJOT/ CELMAJ(1:NDBASE,1:NDBASE,0:NDREVE)
       COMMON &
              /SPNUMS/ NUMBSP(0:NDREVE,0:NDISOS)
       COMMON &
              /DIMENS/ LDBASE
       COMMON &
              /MAXXYM/ NXMAXX,NYMAXX,NZMAXX
       COMMON &
              /T_PHAS/ IPHAPP(0:NDYMAX,0:NDYMAX,0:NDKART), &
                       IPHAPM(0:NDYMAX,0:NDYMAX,0:NDKART), &
                       IPHAMP(0:NDYMAX,0:NDYMAX,0:NDKART), &
                       IPHAMM(0:NDYMAX,0:NDYMAX,0:NDKART)

       C_ZERO=CMPLX(0.0D0,0.0D0)

       ! Number of states
       LSTATE=NUMBSP(0,ICHARG)+MREVER*NUMBSP(1,ICHARG)

       CAUXIL(:,:)=C_ZERO
       FACTOR=1

       ! X-direction
       IF (INDEXJ.EQ.1) THEN

           ISIPAR=-1

           ! IMAREA=-1
           CALL INT_LX(NXMAXX,NYMAXX,NZMAXX,IPHAPM(0,0,0),RAUXIL,FACTOR)
           CALL PUTIMA(+1,RAUXIL,CAUXIL)

           ! IMAREA=+1
           CALL INT_SX(NXMAXX,NYMAXX,NZMAXX,RAUXIL,FACTOR)
           CALL PUTREA(+1,RAUXIL,CAUXIL)

           CALL JXOPER(CAUXIL,CELMAJ,ICHARG)

       END IF

       ! Y-direction
       IF (INDEXJ.EQ.2) THEN

           ISIPAR=+1

           ! IMAREA=-1
           CALL INT_LY(NXMAXX,NYMAXX,NZMAXX,IPHAPP(0,0,0),RAUXIL,FACTOR)
           CALL PUTIMA(+1,RAUXIL,CAUXIL)

           ! IMAREA=+1
           CALL INT_SY(NXMAXX,NYMAXX,NZMAXX,RAUXIL,FACTOR)
           CALL PUTREA(+1,RAUXIL,CAUXIL)

           CALL JYOPER(CAUXIL,CELMAJ,ICHARG)

       END IF

       !  Z-direction
       IF (INDEXJ.EQ.3) THEN

           ISIPAR=-1

           ! IMAREA=-1
           CALL INT_LZ(NXMAXX,NYMAXX,NZMAXX,IPHAPM(0,0,0),RAUXIL,FACTOR)
           CALL PUTIMA(+1,RAUXIL,CAUXIL)

           CALL INT_SZ(NXMAXX,NYMAXX,NZMAXX,RAUXIL,FACTOR)
           CALL PUTIMA(+1,RAUXIL,CAUXIL)

           CALL JXOPER(CAUXIL,CELMAJ,ICHARG)

       END IF

       RETURN
     END SUBROUTINE JMATEL

     !--------------------------------------------------------------------!
     !  This routine calculates the matrix elements <alpha.beta.F>, where !
     !  alpha and beta are qp-annihilation operators and F has the same   !
     !  symmetry properties as J_x or J_z in the good y-simplex basis of  !
     !  HFODD (time-odd, simplex-odd, hermitian). This calculation is     !
     !  done at T=0, meaning only the component F20 is non-zero (F11=0)   !
     !--------------------------------------------------------------------!
     SUBROUTINE JXOPER(SPOPER,F2ELEM,ICHARG)

       USE ALLQUA

       COMPLEX :: C_ZERO,C_UNIT,F2ELEM,SPOPER,DUMMY1,DUMMY2

       DIMENSION :: SPOPER(1:NDBASE,1:NDBASE), &
                    F2ELEM(1:NDBASE,1:NDBASE,0:NDREVE), &
                    DUMMY1(1:NDBASE,1:NDBASE), &
                    DUMMY2(1:NDBASE,1:NDBASE)

       COMMON &
              /QPNUMS/ NUMBQP(0:NDREVE,0:NDISOS)
       COMMON &
              /SPNUMS/ NUMBSP(0:NDREVE,0:NDISOS)
       COMMON &
              /DIMENS/ LDBASE

       C_ZERO=CMPLX(0.0D0,0.0D0)
       C_UNIT=CMPLX(1.0D0,0.0D0)

       F2ELEM = C_ZERO
       DUMMY1 = C_ZERO
       DUMMY2 = C_ZERO

       ! (+,+) block in simplex
       CALL ZGEMM('T','T',NUMBQP(1,ICHARG),LDBASE,LDBASE,    &
                  C_UNIT,AWAQUA(1,1,1),NDBASE,SPOPER,NDBASE, &
                  C_ZERO,DUMMY1,NDBASE)

       CALL ZGEMM('N','N',NUMBQP(1,ICHARG),NUMBQP(0,ICHARG),LDBASE, &
                  C_UNIT,DUMMY1,NDBASE,BWAQUA(1,1,0),NDBASE,        &
                  C_ZERO,DUMMY2,NDBASE)

       DUMMY1 = TRANSPOSE(DUMMY2)
       F2ELEM(:,:,0) = DUMMY1-DUMMY2

       ! (-,-) block in simplex
       DUMMY1 = C_ZERO
       DUMMY2 = C_ZERO

       CALL ZGEMM('T','N',NUMBQP(0,ICHARG),LDBASE,LDBASE,    &
                  C_UNIT,AWAQUA(1,1,0),NDBASE,SPOPER,NDBASE, &
                  C_ZERO,DUMMY1,NDBASE)

       CALL ZGEMM('N','N',NUMBQP(0,ICHARG),NUMBQP(1,ICHARG),LDBASE, &
                  C_UNIT,DUMMY1,NDBASE,BWAQUA(1,1,1),NDBASE,        &
                  C_ZERO,DUMMY2,NDBASE)

       DUMMY1 = TRANSPOSE(DUMMY2)
       F2ELEM(:,:,1) = DUMMY1-DUMMY2

       RETURN
     END SUBROUTINE JXOPER

     !--------------------------------------------------------------------!
     !  This routine calculates the matrix elements <alpha.beta.F>, where !
     !  alpha and beta are qp-annihilation operators and F has the same   !
     !  symmetry properties as J_y in the good y-simplex basis of HFODD   !
     !  (time-odd, simplex-even, hermitian) This calculation is done at   !
     !  T=0, meaning only the component F20 is non-zero (F11=0)           !
     !--------------------------------------------------------------------!
     SUBROUTINE JYOPER(SPOPER,F2ELEM,ICHARG)

       USE ALLQUA

       COMPLEX :: C_ZERO,C_UNIT,F2ELEM,SPOPER,DUMMY1,DUMMY2

       DIMENSION :: SPOPER(1:NDBASE,1:NDBASE),          &
                    F2ELEM(1:NDBASE,1:NDBASE,0:NDREVE), &
                    DUMMY1(1:NDBASE,1:NDBASE),          &
                    DUMMY2(1:NDBASE,1:NDBASE)

       COMMON &
              /QPNUMS/ NUMBQP(0:NDREVE,0:NDISOS)
       COMMON &
              /SPNUMS/ NUMBSP(0:NDREVE,0:NDISOS)
       COMMON &
              /DIMENS/ LDBASE

       C_ZERO=CMPLX(0.0D0,0.0D0)
       C_UNIT=CMPLX(1.0D0,0.0D0)

       F2ELEM = C_ZERO
       DUMMY1 = C_ZERO
       DUMMY2 = C_ZERO

       CALL ZGEMM('T','T',NUMBQP(1,ICHARG),LDBASE,LDBASE,    &
                  C_UNIT,AWAQUA(1,1,1),NDBASE,SPOPER,NDBASE, &
                  C_ZERO,DUMMY1,NDBASE)

       CALL ZGEMM('N','N',NUMBQP(1,ICHARG),NUMBQP(1,ICHARG),LDBASE, &
                  C_UNIT,DUMMY1,NDBASE,BWAQUA(1,1,1),NDBASE,        &
                  C_ZERO,DUMMY2,NDBASE)

       ! (+,-)
       F2ELEM(:,:,0) = DUMMY2
       ! (-,+)
       F2ELEM(:,:,1) = -TRANSPOSE(DUMMY2)

       DUMMY1 = C_ZERO
       DUMMY2 = C_ZERO

       CALL ZGEMM('T','T',NUMBQP(0,ICHARG),LDBASE,LDBASE,    &
                  C_UNIT,BWAQUA(1,1,0),NDBASE,SPOPER,NDBASE, &
                  C_ZERO,DUMMY1,NDBASE)

       CALL ZGEMM('N','N',NUMBQP(0,ICHARG),NUMBQP(0,ICHARG),LDBASE, &
                  C_UNIT,DUMMY1,NDBASE,AWAQUA(1,1,0),NDBASE,        &
                  C_ZERO,DUMMY2,NDBASE)

       ! (+,-) block in simplex
       F2ELEM(:,:,0) = F2ELEM(:,:,0)+DUMMY2

       ! (-,+) block in simplex
       DUMMY1=TRANSPOSE(DUMMY2)
       F2ELEM(:,:,1) = F2ELEM(:,:,1)-DUMMY1

       RETURN
     END SUBROUTINE JYOPER

     !####################   END OF BLOCK ROTCOR ############################


     !-------------------------------------------------------------------!
     !                                                                   !
     !  This subroutine writes data (matrices AWAQUA, BWAQUA and ENQUAP) !
     !  to disk with the same name as HFODD "FILREC" and extension .qp   !
     !                                                                   !
     !  The routine should be called in the main part of the HFODD code  !
     !  at the end of the neutron and proton sections.                   !
     !                                                                   !
     !-------------------------------------------------------------------!
     subroutine writqp(termnt,icharg,mrever,nudata,IF2FIN,IF2FIP,&
                                   fe2fin,fe2fip,energy,e0,INDJOB,name_file)

       USE ALLWAV
       USE ALLQUA

       logical :: TERMNT, name_file
       character(len =  6) :: filrec
       character(len = 19) :: filnam, stringa
       character(len = 68) :: rootname, filqp
       COMPLEX :: C_ZERO,C_UNIT

       COMMON &
              /QPNUMS/ NUMBQP(0:NDREVE,0:NDISOS)
       COMMON &
              /DIMENS/ LDBASE
       COMMON                                &
              /ALLENE/ EKIN_N,EKIN_P,EKIN_T, &
                       EPOT_N,EPOT_P,EPOT_T, &
                       ESUM_N,ESUM_P,ESUM_T, &
                       EPAI_N,EPAI_P,EPAI_T, &
                       EREA_N,EREA_P,EREA_T, &
                       ELIP_N,ELIP_P,ELIP_T, &
                       ECOULD,ECOULE,ECOULT, &
                              ECOULS,ECOULV, &
                       EMULCO,EMUSLO,EMUREA, &
                       ESIFCO,ESISLO,ESIREA, &
                       ESPICO,ESPSLO,ESPREA, &
                       ENREAR,ECORCM,ECOR_R, &
                       EEVEW0,EODDW0,ENE_W0, &
                       ENEVEN,ENEODD,ENESKY, &
                       ESTABN,ETOTSP,ETOTFU
       COMMON                                             &
              /DATPAN/ GPAIRN,FACTGN,EFERMN,DELTAN,EFER2N &
              /DATPAP/ GPAIRP,FACTGP,EFERMP,DELTAP,EFER2P
       COMMON                                           &
              /QCNSTR/ STIFFQ(0:NDMULT,-NDMULT:NDMULT), &
                       QASKED(0:NDMULT,-NDMULT:NDMULT), &
                       IFLAGQ(0:NDMULT,-NDMULT:NDMULT)
       COMMON                                           &
              /QROTAT/ QROT_N(0:NDMULT,-NDMULT:NDMULT), &
                       QROT_P(0:NDMULT,-NDMULT:NDMULT), &
                       QROT_T(0:NDMULT,-NDMULT:NDMULT)
       COMMON                                           &
              /QMULTI/ QMUL_N(0:NDMULT,-NDMULT:NDMULT), &
                       QMUL_P(0:NDMULT,-NDMULT:NDMULT), &
                       QMUL_T(0:NDMULT,-NDMULT:NDMULT)
       COMMON                                     &
              /QUAPAR/ ENQUAP(1:NDSTAT,0:NDREVE), &
                       V2QUAP(1:NDSTAT,0:NDREVE)

       COMMON /CFIPRI/ NFIPRI

       ! iwunit has to be known at ICHARG=0, and 1
       COMMON /SUNITS/ IWUNIT

       if ( .not. TERMNT ) return
       if (IF2FIN.eq.1.or.IF2FIP.eq.1) IIPAIR=1

       if (icharg.eq.0) then

           If (name_file) Then

             number1 = int(qasked(2,0)+0.0001)
             number2 = 999
             number3 = 999
             stringa = 'q00.-'

             if(IFLAGQ(2,2).eq.1)then
                number2 = int(abs(qasked(2,2))+0.0001)
                if(qasked(2,2).lt.0.0d0) number2 =-number2
                stringa = 'q22.'
                if (number2.lt.0) then
                    stringa = 'q22.-'
                    number2 = abs(number2)
                end if
             else if(IFLAGQ(3,0).eq.1)then
                number2 = int(abs(qasked(3,0))+0.0001)
                if(qasked(3,0).lt.0.0d0) number2 =-number2
                stringa = 'q30.'
                if (number2.lt.0) then
                    stringa = 'q30.-'
                    number2 = abs(number2)
                end if
             end if

             if(IIPAIR.eq.1)then
                write(nfipri,'(''lambda2='',2f6.3)') fe2fin,fe2fip
                grid0=(fe2fin+fe2fip)/2.0d0
                number3 = int(abs(grid0)*100.0d0+0.0001)
                if(grid0.lt.0.0d0) number3 =-number3
             endif

             if (number2.lt.0) then
                 write(filnam,'("q20.",I3.3,"_",A5,I3.3,"_l.",I3.3)') number1,Trim(stringa),number2,number3
             else
                 write(filnam,'("q20.",I3.3,"_",A4,I3.3,"_l.",I3.3)') number1,Trim(stringa),number2,number3
             end if

          Else

              write(filnam,'("HFODD_",i8.8)') INDJOB

          End If

          WRITE(FILQP,'(A19,".QP")') Trim(FILNAM)
          CALL GET_UNIT(IWUNIT)
          OPEN(IWUNIT,IOSTAT=ierr,FILE=TRIM(FILQP),STATUS='UNKNOWN',FORM='UNFORMATTED')
          If (ierr.Ne.0) Then
              Write(6,'("Error opening file ",a19)') filnam
              Write(6,'("In write_qp: file will not be recorded on disk")')
              Return
          End If

       END IF

       qrot20=qrot_t(2,0)
       qrot22=qrot_t(2,2)
       qrot30=qrot_t(3,0)

       write(iwunit) qrot20, qrot22, qrot30, grid0*100.0d0, energy, e0, deltan, deltap,&
                     nudata, icharg, numbqp(0,icharg), numbqp(1,icharg), ldbase

       write(iwunit) ((awaqua(iket,jstat,0),iket=1,ldbase),jstat=1,numbqp(0,icharg))

       write(iwunit) ((awaqua(iket,jstat,1),iket=1,ldbase),jstat=1,numbqp(1,icharg))

       write(iwunit) ((bwaqua(iket,jstat,0),iket=1,ldbase),jstat=1,numbqp(0,icharg))

       write(iwunit) ((bwaqua(iket,jstat,1),iket=1,ldbase),jstat=1,numbqp(1,icharg))

       write(iwunit) (enquap(jstat,0),jstat=1,numbqp(0,icharg))

       write(iwunit) (enquap(jstat,1),jstat=1,numbqp(1,icharg))

       ! closing files
       if (icharg.eq.1) then
          close(iwunit)
       end if

       return
     end subroutine writqp

     !-------------------------------------------------------------------!
     !  make_filename creates a unique root file name of input and       !
     !  output files (without extension) based on the three numbers      !
     !  number1, number2, and number 3.                                  !
     !     rootname = prefix                                             !
     !     number   = -99..999                                           !
     !     filename = e.g., prefix-109, prefix--03, etc.                 !
     !-------------------------------------------------------------------!
     Subroutine make_filename(rootname, number1, number2, number3, filename)

       Implicit None

       Integer :: number1, number2, number3
       Character(Len=3) :: string1, string2, string3
       Character(Len=*) :: rootname, filename

       Write(string1,"(i3)") number1
       Write(string2,"(i3)") number2
       Write(string3,"(i3)") number3

       If ( number1.lt.100 ) string1(1:1)="0"
       If ( Abs(number1).lt. 10 ) string1(2:2)="0"
       If ( number1.lt.  0 ) string1(1:1)="-"
       If ( number2.lt.100 ) string2(1:1)="0"
       If ( Abs(number2).lt. 10 ) string2(2:2)="0"
       If ( number2.lt.  0 ) string2(1:1)="-"
       If ( number3.lt.100 ) string3(1:1)="0"
       If ( Abs(number3).lt. 10 ) string3(2:2)="0"
       If ( number3.lt.  0 ) string3(1:1)="-"

       If (number2.ne.-99.and.number3.ne.-99) Then
           filename=trim(rootname)//string1//'-'//string2//'-'//string3
       Else If(number2.eq.-99.and.number3.ne.-99) Then
           filename=trim(rootname)//string1//'-'//string3
       Else If(number2.ne.-99.and.number3.eq.-99) Then
           filename=trim(rootname)//string1//'-'//string2
       Else
           filename=trim(rootname)//string1
       End If

       Return
     End Subroutine make_filename

     !-------------------------------------------------------------------!
     !
     ! GET_UNIT returns a free FORTRAN unit number.
     !
     !  Discussion:
     !
     !    A "free" FORTRAN unit number is an integer between 1 and 99 which
     !    is not currently associated with an I/O device.  A free FORTRAN unit
     !    number is needed in order to open a file with the OPEN command.
     !
     !    If IUNIT = 0, then no free FORTRAN unit could be found, although
     !    all 99 units were checked (except for units 5, 6 and 9, which
     !    are commonly reserved for console I/O).
     !
     !    Otherwise, IUNIT is an integer between 1 and 99, representing a
     !    free FORTRAN unit.  Note that GET_UNIT assumes that units 5 and 6
     !    are special, and will never return those values.
     !
     !  Author:
     !
     !    John Burkardt
     !
     !  Parameters:
     !
     !    Output, integer IUNIT, the free unit number.
     !-------------------------------------------------------------------!
     Subroutine get_unit ( iunit )

       Implicit None

       Logical :: lopen
       Integer :: i,ios,iunit

       iunit = 0

       Do i = 1,99
          If ( i /= 5 .and. i /= 6 .and. i /= 9 ) Then

             Inquire ( unit = i, opened = lopen, iostat = ios )

             If ( ios == 0 ) Then
                If ( .not. lopen ) Then
                   iunit = i
                   Return
                End If
             End If

          End If

       End Do

       Return
     End Subroutine get_unit

END MODULE hfodd_inertia

